#ifndef SU2_H
#define SU2_H

#include <complex>
#include "init.h"

using namespace std;

class SU2 {
	//------------------------------------------
	// SU(2) matrix is represented as a 4-vector
	// U=a_0*I+i\vec{a}*\vec{\tau}, where \tau are Pauli matrices
	// \sum_{\mu}a_{\mu}^2=1 
	//------------------------------------------
	 friend std::ostream &operator<<( std::ostream & out, const SU2 & U){
	 	// Print an SU(2) matrix// change to GetMatrix
	 	std::complex<double> I(0,1);
		out<<"["<<U[0]+I*U[3]<<" "<<I*U[1]+U[2]<<
			std::endl;
		out<<I*U[1]-U[2]<<" "<<U[0]-I*U[3]<<"]";
		return out;
	}
	public:
		SU2() {}
		explicit SU2(double a0, double a1=0, double a2=0, double a3=0)
		{
			SetVector(a0,a1,a2,a3);
		}
		SU2(const SU2 & In);
		void operator()(double a0, double a1=0, double a2=0, double a3=0);
		void SetMatrix(std::complex<double> a00, std::complex<double> a01, 
			std::complex<double> a10, std::complex<double> a11);
		std::complex<double> GetMatrix(int i, int j) const;
		void SetVector(double a0, double a1, double a2, double a3);
		double * GetVector();
		
		double operator[] (const int i) const ;
		SU2 operator*(const SU2& rhs);
		void operator*=(const SU2& rhs);
		void operator*=(const double rhs);
		void operator+=(const SU2& rhs);
		void operator-=(const SU2& rhs);
		
		double trace();
		double Trr(int rep2);
		template<int rep2>
		inline double Tr();
		SU2 dagger();
		SU2 reunit();
	private:
		double a[4];
};
//=========== definitions =====================================================
inline SU2::SU2(const SU2 & In)
{
	SetVector(In[0],In[1],In[2],In[3]);
}
inline void SU2::operator()(double a0, double a1, double a2, double a3) 
{
	SetVector(a0,a1,a2,a3);
}
inline void SU2::SetMatrix(std::complex<double> a00, std::complex<double> a01, 
	std::complex<double> a10, std::complex<double> a11) 
{
	// is not used, we deal only with a vector representation
	a[0]=0.5*(a00.real()+a11.real());
	a[1]=0.5*(a01.imag()+a10.imag());
	a[2]=0.5*(a01.real()-a10.real());
	a[3]=0.5*(a00.imag()-a11.imag());
}
inline std::complex<double> SU2::GetMatrix(int i, int j) const
{
	std::complex<double> I(0,1);
	switch (i) {
		case 0:
			switch (j) {
				case 0:
					return a[0]+I*a[3]; //(0,0)
				case 1:
					return I*a[1]+a[2];//(0,1)
				default:
					std::cerr<<"ERROR: SU2 index larger than 2"<<std::endl;
					abort();
			}
		case 1:
			switch (j) {
				case 0:
					return I*a[1]-a[2]; //(1,0)
				case 1:
					return a[0]-I*a[3];//(1,1)
				default:
					std::cerr<<"ERROR: SU2 index larger than 2"<<std::endl;
					abort();
			}
		default:
			std::cerr<<"ERROR: SU2 index larger than 2"<<std::endl;
			abort();
	}
}
inline void SU2::SetVector(double a0, double a1, double a2, double a3) 
{
	// Set from a vector in a Pauli matrix basis 
	a[0]=a0;
	a[1]=a1;
	a[2]=a2;
	a[3]=a3;
}
inline double * SU2::GetVector() 
{
	return a;
}
inline double SU2::operator[] (const int i) const 
{ 
	return a[i]; 
}
inline SU2 SU2::operator*(const SU2& rhs)
{
	SU2 Temp((*this)[0]*rhs[0]-(*this)[1]*rhs[1]-(*this)[2]*rhs[2]-(*this)[3]*rhs[3],
		(*this)[1]*rhs[0]+(*this)[0]*rhs[1]-(*this)[2]*rhs[3]+(*this)[3]*rhs[2],
		(*this)[2]*rhs[0]+(*this)[0]*rhs[2]-(*this)[3]*rhs[1]+(*this)[1]*rhs[3],
		(*this)[3]*rhs[0]+(*this)[0]*rhs[3]-(*this)[1]*rhs[2]+(*this)[2]*rhs[1]);
	return Temp;
}
inline void SU2::operator*=(const SU2& rhs)
{
	double b0=(*this)[0]*rhs[0]-(*this)[1]*rhs[1]-(*this)[2]*rhs[2]-(*this)[3]*rhs[3];
	double b1=(*this)[1]*rhs[0]+(*this)[0]*rhs[1]-(*this)[2]*rhs[3]+(*this)[3]*rhs[2];
	double b2=(*this)[2]*rhs[0]+(*this)[0]*rhs[2]-(*this)[3]*rhs[1]+(*this)[1]*rhs[3];
	double b3=(*this)[3]*rhs[0]+(*this)[0]*rhs[3]-(*this)[1]*rhs[2]+(*this)[2]*rhs[1];
	SetVector(b0,b1,b2,b3);
}
inline void SU2::operator*=(const double rhs)
{
	SetVector((*this)[0]*rhs,(*this)[1]*rhs,
		(*this)[2]*rhs,(*this)[3]*rhs);
}
inline void SU2::operator+=(const SU2& rhs)
{
	SetVector((*this)[0]+rhs[0],(*this)[1]+rhs[1],
		(*this)[2]+rhs[2],(*this)[3]+rhs[3]);
}
inline void SU2::operator-=(const SU2& rhs)
{
	SetVector((*this)[0]-rhs[0],(*this)[1]-rhs[1],
		(*this)[2]-rhs[2],(*this)[3]-rhs[3]);
}
inline double SU2::trace()//SU2 only - trace is real
{
	return 2*a[0];
}
inline double SU2::Trr(int rep2)
{//representation =rep2/2
	double epsilon=1e-19;
	//if(a[0]>1) cout<<endl<<"a0 "<<a[0]<<endl;
//	assert(a[0]<1+1e-5 && a[0]>-1-1e-5);
	//if(a[0]>1) (*this).reunit(); //remove rounding problem
	if(a[0]>1||a[0]<-1) {
		cout<<"WARNING: had to reunitarize in trace"<<endl;
		cout<<"a0="<<a[0]<<endl;
		//remove rounding problem
		double detsqrt=sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]+a[3]*a[3]);
		a[0]/=detsqrt;
		a[1]/=detsqrt;
		a[2]/=detsqrt;
		a[3]/=detsqrt;
	}
	double theta2=acos(a[0]);
	//cout<<"t:"<<a[0]<<" "<<theta2<<endl;
	if(theta2<epsilon) return (rep2+1)*pow(a[0],rep2);
	if(theta2>M_PI-epsilon) return (rep2+1)*pow(a[0],rep2);
	return sin((rep2+1.0)*theta2)/sin(theta2);
}
template<int rep2>
inline double SU2::Tr()
{//representation =rep2/2
	double epsilon=1e-29;
	assert(a[0]<=1);
	double theta2=acos(a[0]);
	if(theta2<epsilon) return (rep2+1)*pow(a[0],rep2);
	return sin((rep2+1.0)*theta2)/sin(theta2);
}
inline SU2 SU2::dagger()
{
	SU2 Temp((*this)[0],-(*this)[1],-(*this)[2],-(*this)[3]);
	return Temp;
}
inline SU2 SU2::reunit()
{
	double det_sqrt=sqrt((*this)[0]*(*this)[0]+(*this)[1]*(*this)[1]+
		(*this)[2]*(*this)[2]+(*this)[3]*(*this)[3]);
	SU2 Temp((*this)[0]/det_sqrt,(*this)[1]/det_sqrt,
		(*this)[2]/det_sqrt,(*this)[3]/det_sqrt);
	return Temp;
}
#endif
