// so far only 4d: Lx Ly Lz Lt lattices with periodic bc
#ifndef LATTICE_H
#define LATTICE_H

#include "init.h"
#include "parameters.h"
#include "SU2.h"
#include "random.h"

class Lattice {
	friend class su2hb;
	friend class Meas;
	friend class SU2Wilson;
	friend class SU2Mixed;
	friend class Demon;
	friend class DemonMixed;
public:
	Lattice(const Params& par); //doesn't initialize the field
	Lattice(const Lattice& InLat);
	//Init the gauge field to unit, random or predefined SU(N) matrices 
	//or read the configuration from a file (depending on State parameter) 
	void Init(enum State);
	void Init(Lattice &);
	SU2 *Umu(const int LinCoord, const int mu);
	SU2 *UmuDec(const int LinCoord, const int mu);
	void Load(const char *);
	void Unload(const char *);
	void DecUnload(const char *);
	void Decimate(int DecType); //1- Swendsen, 2- Double Smeared
	SU2 Staples(const int LinCoord, const int mu);
	SU2 Staples(const int LinCoord, const int mu, SU2 (*Uf)[Nd]);
	SU2 Staple(const int LinCoord, const int mu, const int nu,const int dir);
	SU2 LongStaples(const int LinCoord, const int mu);
	SU2 LongStaplesDagger(const int LinCoord, const int mu);
	inline SU2 LongLink(const int LinCoord, const int mu,int Length=2);
	inline SU2 LongLink(const int LinCoord, const int mu,SU2 (*Uf)[Nd],int Length=2);
	inline int CoordShift(const int LinCoord, const LatVec& ShiftVec);
	inline void Lin2Vec(const int LinCoord, LatVec & LatCoord);
	inline void Vec2Lin(const LatVec LatCoord, int &LinCoord);
	inline void LinDec2Vec(const int LinCoord, LatVec & LatCoord);
	~Lattice();

private:
	void DecimateSW();//Swendsen
	void DecimateDS();//Double smeared blocking
	void RandomSphere(const double rad, double &a1,double &a2,double &a3);
	Random * prnd;
	const int lx, ly, lz, lt; //lattice size (Nd=4 is hard coded)
	const int Vol; //volume
	const int Nrep; //total number of used representations
	SU2 (*Ufield)[Nd]; //gauge field on links
	SU2 (*UfDec)[Nd]; //pointer to gauge field on decimated links
	
	vector<double> Beta;
	double Betap;
	double CDec;
};

//=========== definitions =====================================================
inline SU2 * Lattice::Umu(const int LinCoord, const int mu) {
	return &(Ufield[LinCoord][mu]);
}
inline SU2 * Lattice::UmuDec(const int LinCoord, const int mu) {
	return &(UfDec[LinCoord][mu]);
}
int Lattice::CoordShift(const int LinCoord, const LatVec& ShiftVec)
{
	LatVec LatCoord;
	int NewLinCoord;
	Lattice::Lin2Vec(LinCoord, LatCoord);
	LatCoord[0]+=ShiftVec[0];
	LatCoord[1]+=ShiftVec[1];
	LatCoord[2]+=ShiftVec[2];
	LatCoord[3]+=ShiftVec[3];
	Vec2Lin(LatCoord,NewLinCoord);
	return NewLinCoord;
}
void Lattice::Lin2Vec(const int LinCoord, LatVec & LatCoord)
{// periodicity is hidden here
	int i=LinCoord;
	LatCoord[3]=i%lt;
	i/=lt;
	LatCoord[0]=i%lx;
	i/=lx;
	LatCoord[1]=i%ly;
	LatCoord[2]=i/ly;
}
void Lattice::Vec2Lin(const LatVec LatCoord, int &LinCoord)
{
	// Periodic boundary condition 
	int rx=(LatCoord[0]+lx)%lx;
	int ry=(LatCoord[1]+ly)%ly;
	int rz=(LatCoord[2]+lz)%lz;
	int rt=(LatCoord[3]+lt)%lt;
	LinCoord=rt+lt*(rx+lx*(ry+ly*rz));
}
void Lattice::LinDec2Vec(const int LinCoord, LatVec & LatCoord)
{// periodicity is hidden here
	int i=LinCoord;
	int dlx=lx/2,dly=ly/2,dlt=lt/2;
	LatCoord[3]=i%dlt;
	i/=dlt;
	LatCoord[0]=i%dlx;
	i/=dlx;
	LatCoord[1]=i%dly;
	LatCoord[2]=i/dly;
}
SU2 Lattice::LongLink(const int lcrd1, const int mu, int Length) 
{
	SU2 LU(Ufield[lcrd1][mu]);
	int lcrd2;
	LatVec crd;
	Lin2Vec(lcrd1,crd);
	for(int i=1;i<Length;i++) {
		crd[mu]+=1;
		Vec2Lin(crd,lcrd2);
		LU*=Ufield[lcrd2][mu];
	}

	return LU;
}
SU2 Lattice::LongLink(const int lcrd1, const int mu, SU2 (*Uf)[Nd], int Length) 
{
	SU2 LU(Uf[lcrd1][mu]);
	int lcrd2;
	LatVec crd;
	Lin2Vec(lcrd1,crd);
	for(int i=1;i<Length;i++) {
		crd[mu]+=1;
		Vec2Lin(crd,lcrd2);
		LU*=Uf[lcrd2][mu];
	}

	return LU;
}
#endif
