#include "SU2Wilson.h"
#define CREUTZ
SU2Wilson::SU2Wilson(Lattice &MyLat)
{
	Lat=&MyLat;
#ifdef CREUTZ	
	cout<<"Using Creutz Heat-Bath."<<endl;
#else
	cout<<"Using Kennedy-Pendleton Heat-Bath."<<endl;
#endif
}

#ifdef CREUTZ
void SU2Wilson::Creutz(const double weight, double &a0)
{
	double w_exp=exp(-2.0*weight);
	double xr=1;
	a0=1.0;
	while(xr >= 1-a0*a0) {
		xr=Lat->prnd->get();
		a0=1.0+log(w_exp*(1-xr)+xr)/weight;
		xr=Lat->prnd->get();
		xr=xr*xr;
	}
}
#else
void SU2Wilson::KennedyPendleton(const double weight, double &a0)
{
	bool accept=false;
	while(!accept) {
		double xr1=Lat->prnd->get();
		double xr2=Lat->prnd->get();
		double xr3=Lat->prnd->get();
		double xr4=Lat->prnd->get();

		xr1=-(log(xr1)/weight);
		xr3 = cos(2.0l*M_PI*xr3);
		xr3=xr3*xr3;
		xr2=-(log(xr2)/weight);
		a0=xr2+xr1*xr3;
		accept=((xr4*xr4) < (1.0-a0/2.0));
	}
	a0=1-a0;
}
#endif

void SU2Wilson::Evolve(int NmOR)
{
	if(Lat->Betap!=Lat->Beta[0]) {
		cout<<"\nWARNING: Betap="<<Lat->Betap<<" is different from Beta[0] ";
		cout<<"for fundamental only updates.\n";
		Lat->Betap=Lat->Beta[0];
		cout<<"Adjusting: Betap=Beta[0]="<<Lat->Betap<<endl;
//		cerr<<"adjust betap to beta[0] in parameter file"<<endl;
//		abort();
	}
	
	for (int LinCoord=0; LinCoord<Lat->Vol; LinCoord++) {
		for (int mu=0;mu<Nd;mu++) {
			*Lat->Umu(LinCoord,  mu)=HBtry(LinCoord,mu);
		}
	}
// Overrelaxation block
//#define SU2WILSON_EVOLVE_MOR_ACTDIFF
	for (int iOR=0; iOR<NmOR; iOR++) {
#ifdef SU2WILSON_EVOLVE_MOR_ACTDIFF
		double actDiff=0;
		double actOld,actNew;
#endif // SU2WILSON_EVOLVE_MOR_ACTDIFF
		for (int LinCoord=0; LinCoord<Lat->Vol; LinCoord++) {
			for (int mu=0;mu<Nd;mu++) {
#ifdef SU2WILSON_EVOLVE_MOR_ACTDIFF
				// calculate old action
				actOld=Lat->Betap*
				  (1-(*Lat->Umu(LinCoord,  mu)*Lat->Staples(LinCoord,mu)).trace()/2);
#endif // SU2WILSON_EVOLVE_MOR_ACTDIFF

				*Lat->Umu(LinCoord,  mu)=ORtry(LinCoord,mu);
#ifdef SU2WILSON_EVOLVE_MOR_ACTDIFF
				// calculate new action
				actNew=Lat->Betap*
				  (1-(*Lat->Umu(LinCoord,  mu)*Lat->Staples(LinCoord,mu)).trace()/2);
				actDiff+=fabs(actNew-actOld);
#endif // SU2WILSON_EVOLVE_MOR_ACTDIFF

			}
		}
#ifdef SU2WILSON_EVOLVE_MOR_ACTDIFF
		cerr<<"mOR update "<<iOR<<" action error: "<<actDiff<<endl;
#endif // SU2WILSON_EVOLVE_MOR_ACTDIFF
	}
}

void SU2Wilson::EvolveSlice(int tin, int tfin)
{
	LatVec crd;
	int lcrd;
	for(crd[2]=0;crd[2]<Lat->lz;crd[2]++) {
		for(crd[1]=0;crd[1]<Lat->ly;crd[1]++) {
			for(crd[0]=0;crd[0]<Lat->lx;crd[0]++) {
				for(crd[3]=tin+1;crd[3]<tfin;crd[3]++) {
					Lat->Vec2Lin(crd,lcrd);
					for (int mu=0;mu<Nd;mu++) {
						*Lat->Umu(lcrd,  mu)=HBtry(lcrd,mu);
					}
				}
				crd[3]=tin;
				Lat->Vec2Lin(crd,lcrd);
				*Lat->Umu(lcrd,  3)=HBtry(lcrd,3);
				
			}
		}
	}
/*	LatVec LatCoord;// Change to Cartesian coords !!!!!!!!!!!!
	for (int LinCoord=0; LinCoord<Lat->Vol; LinCoord++) {
		Lat->Lin2Vec(LinCoord,LatCoord);
		if(LatCoord[3]>tin && LatCoord[3]<tfin) {
			for (int mu=0;mu<Nd;mu++) {
				*Lat->Umu(LinCoord,  mu)=HBtry(LinCoord,mu);
			}
		}
		else if(LatCoord[3]==tin) {
			*Lat->Umu(LinCoord,  3)=HBtry(LinCoord,3);
		}
	}*/
}

SU2 SU2Wilson::HBtry(int LinCoord, int mu) 
{
	SU2 Vmu;
			//v=u*u_staple
			Vmu=(*Lat->Umu(LinCoord,  mu))*
				Lat->Staples(LinCoord,mu);
			double SqDet=sqrt(Vmu[0]*Vmu[0]+Vmu[1]*Vmu[1]+Vmu[2]*Vmu[2]+Vmu[3]*Vmu[3]);
			Vmu.SetVector(Vmu[0]/SqDet,-Vmu[1]/SqDet,-Vmu[2]/SqDet,-Vmu[3]/SqDet);

			Vec4 aVec;
#ifdef CREUTZ
			Creutz(Lat->Betap*SqDet,aVec[0]);
#else
			KennedyPendleton(Lat->Betap*SqDet,aVec[0]);
#endif

			//other aVec components
			Lat->RandomSphere(sqrt(1-aVec[0]*aVec[0]),aVec[1],aVec[2],aVec[3]);

			SU2 Uhb(aVec[0],aVec[1],aVec[2],aVec[3]);
	return Uhb*Vmu*(*Lat->Umu(LinCoord,  mu));
}

SU2 SU2Wilson::ORtry(int LinCoord, int mu) 
{
	// Use the following speed-up:
        // (let Us=staple, Usn=normalized staple, k2=det|Us|, Usn=Us/sqrt(k2))
        // U_new = Usn^+ U_old^+ Usn^+ = (Usn U_old Usn)^+ =
        //       = (Us U_old Us)^+ / k2
	SU2 Vmu;
			//v=Us
			Vmu=Lat->Staples(LinCoord,mu);
			double k2=Vmu[0]*Vmu[0]+Vmu[1]*Vmu[1]+Vmu[2]*Vmu[2]+Vmu[3]*Vmu[3];

			// v=Us * U_old * Us
			Vmu=Vmu*(*Lat->Umu(LinCoord,mu))*Vmu;

			// v=v^+ / k2
			Vmu.SetVector(Vmu[0]/k2,-Vmu[1]/k2,-Vmu[2]/k2,-Vmu[3]/k2);

	return Vmu;
}

void SU2Wilson::Equilibrium(int EqSweeps)
{
//	if(Lat->MyState != STARTFILE) {
		int Eskip=1;
		if(EqSweeps>10) Eskip=EqSweeps/10;  
		cout<<"Equilibrization: ";
		for (int sweeps=0; sweeps<EqSweeps; sweeps++) {
			if(sweeps%Eskip == 0) 
				cout<<sweeps<<" "<<flush;
			Evolve();
		}
		cout<<EqSweeps<<endl;
//	}
}

SU2Wilson::~SU2Wilson()
{}
