#include "lattice.h"

Lattice::Lattice(const Params& par)
	:lx(par.LatSize[0]), ly(par.LatSize[1]), lz(par.LatSize[2]), lt(par.LatSize[3]), 
	Vol(lx*ly*lz*lt), Nrep(par.Nrep)
{
	if(Nd!=4) {
		std::cerr<< "Only 4d is implemented"<<std::endl;
		exit(-1);
	}
	Ufield = new SU2[Vol][Nd]; //customize new for faster allocation?
	UfDec = new SU2[Vol/16][Nd];
//	UfDec = new SU2[Vol][Nd];
	Beta=par.Beta;
	Betap=par.Betap;
	prnd=new Random(par.getIseed());
	CDec=par.getDecSWc();
}

Lattice::Lattice(const Lattice& InLat)
	:lx(InLat.lx),ly(InLat.ly), lz(InLat.lz), lt(InLat.lt), 
	Vol(InLat.Vol), Nrep(InLat.Nrep)
{
	Ufield = new SU2[Vol][Nd]; 
	UfDec = new SU2[Vol/16][Nd];
	Beta=InLat.Beta;
	Betap=InLat.Betap;
	prnd=InLat.prnd;
	CDec=InLat.CDec;
	for (int i=0;i<Vol;i++) {
		for (int mu=0;mu<Nd;mu++) Ufield[i][mu]=InLat.Ufield[i][mu];
	}
}
void Lattice::Init(enum State MyState)
{
	switch(MyState) {
		case ORDERED: //ordered start
			for (int i=0;i<Vol;i++) {
				for (int mu=0;mu<Nd;mu++) Ufield[i][mu](1.0);
			}
			break;
		case RANDOM: //disordered start
			Vec4 aVec;
			for (int i=0; i<Vol;i++) {
				for (int mu=0;mu<Nd;mu++) {
					aVec[0]=1.0-2*prnd->get();
					RandomSphere(sqrt(1-aVec[0]*aVec[0]),aVec[1],aVec[2],aVec[3]);
					Ufield[i][mu](aVec[0],aVec[1],aVec[2],aVec[3]);
				}
			}
			break;
		case STARTFILE: //start from a file
			Load("last.d");//will be changed
			break;
		default:
			std::cerr << "This state is not defined" << std::endl;
			abort();
	}
}

void Lattice::Init(Lattice & OrigLat) 
{
	for (int i=0;i<Vol;i++) {
		for (int mu=0;mu<Nd;mu++) Ufield[i][mu]=*OrigLat.UmuDec(i,mu);
	}
}

void Lattice::Load(const char *LatFileName)
{
	ifstream inFile(LatFileName, ios::in);
	if(!inFile) {
		std::cerr<<"Cannot open lattice input file"<<std::endl;
		abort();
	}
	//read and check header
	int lx_in, ly_in,lz_in,lt_in;
	inFile>>lx_in>>ly_in>>lz_in>>lt_in;
	if(lx_in!=lx || ly_in!=ly || lz_in!=lz || lt_in!=lt) {
		cerr<<"STARTFILE: header/parameters mismatch"<<endl;
		abort();
	}
	//read the field
	double in_a[4];
	for (int i=0;i<Vol;i++) {
		for (int mu=0;mu<Nd;mu++) {
			for (int j=0;j<4;j++) {
				inFile >> in_a[j];
			}
			Ufield[i][mu].SetVector(in_a[0],in_a[1],in_a[2],in_a[3]);
		}
	}
}

void Lattice::Unload(const char * LatFileName)
{
	ofstream outFile(LatFileName, ios::out);
	if(!outFile) {
		std::cerr<<"Cannot open lattice output file"<<std::endl;
		abort();
	}
	//write the header
	outFile << lx<<" "<<ly<<" "<<lz<<" "<<lt<<endl;
	//write the field configuration
	for (int i=0;i<Vol;i++) {
		for (int mu=0;mu<Nd;mu++) {
			for (int j=0;j<4;j++) 
				outFile << Ufield[i][mu].GetVector()[j]<<" ";
		}
	}
}

void Lattice::DecUnload(const char * LatFileName)
{
	ofstream outFile(LatFileName, ios::out);
	if(!outFile) {
		std::cerr<<"Cannot open lattice output file"<<std::endl;
		abort();
	}
	//write the header
	outFile << lx/2<<" "<<ly/2<<" "<<lz/2<<" "<<lt/2<<endl;
	//write the field configuration
	for (int i=0;i<Vol/16;i++) {
		for (int mu=0;mu<Nd;mu++) {
			for (int j=0;j<4;j++) 
				outFile << UfDec[i][mu].GetVector()[j]<<" ";
		}
	}
}

SU2 Lattice::Staples(const int LinCoord, const int mu)
{
	LatVec ShiftVec,ShiftVec_mu;
	ShiftVec_mu[0]=0;ShiftVec_mu[1]=0;ShiftVec_mu[2]=0;ShiftVec_mu[3]=0;
	ShiftVec_mu[mu]=1;
	int TempCoord;
	SU2 Staples(0);
	for(int nu=0; (nu < Nd); nu++) {
		if(nu==mu) continue;
		ShiftVec[0]=0;ShiftVec[1]=0;ShiftVec[2]=0;ShiftVec[3]=0;ShiftVec[nu]=1;
		Staples+=Ufield[CoordShift(LinCoord, ShiftVec_mu)][nu]*
			Ufield[CoordShift(LinCoord, ShiftVec)][mu].dagger()*
			Ufield[LinCoord][nu].dagger(); // forward staples
		
		ShiftVec[nu]=-1;
		TempCoord=CoordShift(LinCoord, ShiftVec);//shift by -nu
		Staples+=Ufield[CoordShift(TempCoord, ShiftVec_mu)][nu].dagger()*
			Ufield[TempCoord][mu].dagger()*
			Ufield[TempCoord][nu]; // backward staples
	}
	return Staples;
}
SU2 Lattice::Staples(const int LinCoord, const int mu, SU2 (*Uf)[Nd])
{
	LatVec ShiftVec,ShiftVec_mu;
	ShiftVec_mu[0]=0;ShiftVec_mu[1]=0;ShiftVec_mu[2]=0;ShiftVec_mu[3]=0;
	ShiftVec_mu[mu]=1;
	int TempCoord;
	SU2 Staples(0);
	for(int nu=0; (nu < Nd); nu++) {
		if(nu==mu) continue;
		ShiftVec[0]=0;ShiftVec[1]=0;ShiftVec[2]=0;ShiftVec[3]=0;ShiftVec[nu]=1;
		Staples+=Uf[CoordShift(LinCoord, ShiftVec_mu)][nu]*
			Uf[CoordShift(LinCoord, ShiftVec)][mu].dagger()*
			Uf[LinCoord][nu].dagger(); // forward staples
		
		ShiftVec[nu]=-1;
		TempCoord=CoordShift(LinCoord, ShiftVec);//shift by -nu
		Staples+=Uf[CoordShift(TempCoord, ShiftVec_mu)][nu].dagger()*
			Uf[TempCoord][mu].dagger()*
			Uf[TempCoord][nu]; // backward staples
	}
	return Staples;
}

SU2 Lattice::Staple(const int LinCoord, const int mu,const int nu,const int dir)
{//check it later!!!!
	assert(nu!=mu);
	LatVec ShiftVec,ShiftVec_mu;
	ShiftVec_mu[0]=0;ShiftVec_mu[1]=0;ShiftVec_mu[2]=0;ShiftVec_mu[3]=0;
	ShiftVec_mu[mu]=1;
	int TempCoord;
	SU2 Staple;

	if (dir>0) {
		ShiftVec[0]=0;ShiftVec[1]=0;ShiftVec[2]=0;ShiftVec[3]=0;ShiftVec[nu]=1;
		Staple=Ufield[CoordShift(LinCoord, ShiftVec_mu)][nu]*
			Ufield[CoordShift(LinCoord, ShiftVec)][mu].dagger()*
			Ufield[LinCoord][nu].dagger(); // forward staples
		return Staple;
	}
	if(dir<0) {
		ShiftVec[0]=0;ShiftVec[1]=0;ShiftVec[2]=0;ShiftVec[3]=0;ShiftVec[nu]=-1;
		TempCoord=Lattice::CoordShift(LinCoord, ShiftVec);//shift by -nu
		Staple=Ufield[CoordShift(TempCoord,  ShiftVec_mu)][nu].dagger()*
			Ufield[TempCoord][mu].dagger()*
			Ufield[TempCoord][nu]; // backward staples
		return Staple;
	}
	if(dir==0) {
		cerr<<"Wrong staple direction"<<endl;
		abort();
	}
	return Staple; //to keep the compiler happy (never gets here)
}

SU2 Lattice::LongStaples(const int LinCoord, const int mu)
{
	SU2 LStaples(0);
	int TempCoord;
	LatVec ShiftVec_nu,ShiftVec_mu;
	ShiftVec_mu[0]=0;ShiftVec_mu[1]=0;ShiftVec_mu[2]=0;ShiftVec_mu[3]=0;
	ShiftVec_mu[mu]=2;//2 link shift
	for(int nu=0;nu<Nd;nu++) {
		if (nu==mu) continue;
		ShiftVec_nu[0]=0;ShiftVec_nu[1]=0;ShiftVec_nu[2]=0;ShiftVec_nu[3]=0;
		ShiftVec_nu[nu]=1;
		LStaples+=Ufield[CoordShift(LinCoord, ShiftVec_mu)][nu]*
			LongLink(CoordShift(LinCoord,ShiftVec_nu),mu).dagger()*
			Ufield[LinCoord][nu].dagger(); // forward long staples
		ShiftVec_nu[nu]=-1;
		TempCoord=CoordShift(LinCoord, ShiftVec_nu);//shift by -nu
		LStaples+=Ufield[CoordShift(TempCoord, ShiftVec_mu)][nu].dagger()*
			LongLink(TempCoord,mu).dagger()*
			Ufield[TempCoord][nu]; // backward staples
	}
	return LStaples;
}

SU2 Lattice::LongStaplesDagger(const int LinCoord, const int mu)
{
	SU2 LStaples(0);
	int TempCoord;
	LatVec ShiftVec_nu,ShiftVec_mu;
	ShiftVec_mu[0]=0;ShiftVec_mu[1]=0;ShiftVec_mu[2]=0;ShiftVec_mu[3]=0;
	ShiftVec_mu[mu]=2;//2 link shift
	for(int nu=0;nu<Nd;nu++) {
		if (nu==mu) continue;
		ShiftVec_nu[0]=0;ShiftVec_nu[1]=0;ShiftVec_nu[2]=0;ShiftVec_nu[3]=0;
		ShiftVec_nu[nu]=1;
		LStaples+=Ufield[LinCoord][nu]*
			LongLink(CoordShift(LinCoord,ShiftVec_nu),mu)*
			Ufield[CoordShift(LinCoord, ShiftVec_mu)][nu].dagger(); // forward long staples
		ShiftVec_nu[nu]=-1;
		TempCoord=CoordShift(LinCoord, ShiftVec_nu);//shift by -nu
		LStaples+=Ufield[TempCoord][nu].dagger()*
			LongLink(TempCoord,mu)*
			Ufield[CoordShift(TempCoord, ShiftVec_mu)][nu]; // backward staples
	}
	return LStaples;
}

void Lattice::RandomSphere(const double rad, double &a1,double &a2,double &a3)
{
	//other a components
	//probably better to move it to SU2.H
			double  CosTheta, Phi;
			assert(rad>=0);
			CosTheta=prnd->get();
			CosTheta=1.0-2.0*CosTheta;
			a3=rad*CosTheta;
	
			CosTheta=(1-CosTheta*CosTheta);
			CosTheta=sqrt(CosTheta);//SinTheta
			Phi=prnd->get();
			Phi*=2*M_PI;

			a1=rad*CosTheta*cos(Phi);
			a2=rad*CosTheta*sin(Phi);
}
void Lattice::Decimate(int DecType)
{
	if(lx%2+ly%2+lz%2+lt%2!=0) {
		cerr<<
		"Cannot decimate: lattice has to have even number of points in all directions"
		<<endl; 
		abort();
	}
	if(DecType==1) DecimateSW();
	if(DecType==2) DecimateDS();
}

void Lattice::DecimateSW()
{
	int VolDec=Vol/16;//4d only
	
	LatVec crd;	
	SU2 UDec;//one decimated link
	int lcrd1;
	for (int i=0;i<VolDec;i++) {
		LinDec2Vec(i,crd);
	//	cout<<endl<<i<<": "<<crd[2]<<","<<crd[1]<<","<<crd[0]<<","<<crd[3]<<endl;
		for(int mu=0;mu<Nd;mu++) crd[mu]*=2;
	//	cout<<i<<": "<<crd[2]<<","<<crd[1]<<","<<crd[0]<<","<<crd[3]<<endl;
		Vec2Lin(crd,lcrd1);//lcrd1 - linear coord on the original lattie
	//	cout<<lcrd1<<endl;
		for(int mu=0;mu<Nd;mu++) {
			UDec=LongLink(lcrd1, mu);
			//UDec=Ufield[lcrd1][mu];
			
			//now add long staple
			SU2 LStaples;
			//LStaples=LongStaples(lcrd1,mu).dagger();
			LStaples=LongStaplesDagger(lcrd1,mu);
			LStaples*=CDec;
			UDec+=LStaples;
			UfDec[i][mu]=UDec.reunit();
		}
	}
}

void Lattice::DecimateDS()
{
	int VolDec=Vol/16;//4d only
	//double CDecDS=0.077;
	SU2 UDec;//one decimated link
	SU2 (*Usmeared1)[Nd],(*Usmeared2)[Nd];
	Usmeared1 = new SU2[Vol][Nd];
	Usmeared2 = new SU2[Vol][Nd];
	for (int i=0;i<Vol;i++) {
		for(int mu=0;mu<Nd;mu++) {
			Usmeared1[i][mu]=Ufield[i][mu];
			Usmeared1[i][mu]*=(1-6*CDec);
			SU2 Ustaples(Staples(i,mu).dagger());
			Ustaples*=CDec;
			Usmeared1[i][mu]+=Ustaples;
		}
	}
	for (int i=0;i<Vol;i++) {
		for(int mu=0;mu<Nd;mu++) {
			Usmeared2[i][mu]=Usmeared1[i][mu];
			Usmeared2[i][mu]*=(1-6*CDec);
			SU2 Ustaples(Staples(i,mu,Usmeared1).dagger());
			Ustaples*=CDec;
			Usmeared2[i][mu]+=Ustaples;
		}
	}
	delete [] Usmeared1;
	LatVec crd;	
	int lcrd1;
	for (int i=0;i<VolDec;i++) {
		LinDec2Vec(i,crd);
		for(int mu=0;mu<Nd;mu++) crd[mu]*=2;
		Vec2Lin(crd,lcrd1);//lcrd1 - linear coord on the original lattie
		for(int mu=0;mu<Nd;mu++) {
			UDec=LongLink(lcrd1, mu,Usmeared2);
			UfDec[i][mu]=UDec.reunit();
		}

	}
	delete [] Usmeared2;
}

Lattice::~Lattice()
{
	delete [] Ufield;
	delete [] UfDec;
}
